# Microarchitetture

In classe abbiamo visto l'architettura ARMv7, che è l'insieme delle istruzioni che possono essere eseguite su un processore ARM, e la loro traduzione in linguaggio macchina.
La microarchitettura si trova ad un livello di astrazione inferiore: lo stesso codice macchina può essere eseguito su più microarchitetture con livelli di efficienza diversi. Alcuni tipi di microarchitetture sono la single-cycle, la multi-cycle e la pipeline.

## Struttura di una microarchitettura

Il processore esegue lo stesso codice per tutta la vita: fetcha una istruzione dalla cache, poi effettua una operazione di decodifica, e poi va ad eseguire (solitamente questo è il lavoro della alu). Esiste anche una fase di interazione con la memoria, e una fase finale chiama write-back.
Nel processore single-cycle tutte queste fasi devono essere eseguite in un solo ciclo di clock.

Possiamo vedere una microarchitettura come divisa in due parti:
- **Data Path** (parte operativa) che contiene la memoria, il register file, il program counter, la ALU, molti multiplexer per "instradare".
- **Parte controllo** limita l'azione del data path.

La parte controllo decide cosa fa la alu. Il datapath fornisce alcuni bit alla parte controllo, che decide come controllare le risorse della data path, e manda indietro dei comandi.

## Micro-architettura single-cycle

Come abbiamo preannunciato, nel processore single-cycle, tutti cicli fetch-decode-execute devono essere svolte in un singolo ciclo di clock.

**Cos'è lo stato architetturale?**
Ci sono nel data path degli elementi di stato: questi sono il program counter, il register file, e la data memory.
Avremo una memoria istruzioni che prenderà un indirizzo da 32 bit A (address), fornirà una uscita da 32 bit che è l'istruzione. E' una DRAM in sola lettura (in teoria).
La memoria dati la consideriamo separata. La memoria dati avrà l'ingresso A, un segnale di clock, il write enable, e WD (ingresso dati) di 32 bit, che è il dato da scrivere. Avrò una uscita da 32 bit chiamata RD. Anche questa è una DRAM.
Il register file contiene al suo interno fisicamente quindici registri, da r0 a r14. Avrà una porta in lettura, una porta in scrittura con il suo write-enable.
Il register file avrà un segnale di clock, due porte in sola lettura, due indirizzi da quattro bit, A1 e A2, che indicano su quali registri dobbiamo leggere/scrivere. Ha due uscite R1 e R2, che contengono i valori dei due registri, in 32 bit. Ha una porta in scrittura, A3, da quattro bit. Prenderà un campo da scrivere da 32 bit che chiamiamo WD3.
Potremo leggere due registri e scriverne un terzo.
Il program counter è un registro da 32 bit senza write enable, perchè viene scritto ad ogni ciclo di clock.

### Costruzione del data path

Come collegare gli elementi dello stato architetturale con il data path?
**Esempio: istruzione di load.**
``` Assembly
LDR R1, [R2, #OFF]
```
<!memory instruction>()

Innanzitutto avremo il nostro program counter che è un registro, che viene letto e all'inizio del ciclo di clock produrrà un valore a 32 bit che è l'indirizzo dell'istruzione da fetchare in memoria.
Questo indirizzo va alla memoria istruzioni. Dalla memoria istruzioni avremo un'uscita di 32 bit, ovvero l'istruzione. Noi dobbiamo leggere il registro R2. Avremo bisogno dei bit Rn. Questi quattro bit vengono passati in ingresso alla porta A1, e avremo nella prima uscita il contenuto del registro due.
Dobbiamo prendere anche i 12 bit che fanno parte del campo dell'immediato offset.
L'offset deve essere esteso, attraverso un estensore, che è una rete combinatoria che prende un immediato di 12 bit e ne fa uscire uno da 32.
Avremmo bisogno di una ALU perchè faccia la somma tra il contenuto di R2 e l'immediato. Fatto questo, otteniamo l'indirizzo della load. Questo indirizzo viene passato alla memoria dati, che ha in uscita 32 bit, il valore che abbiamo caricato.
L'uscita della memoria dati viene portata al campo WD3 del register file. Nel campo A3 del register file mettiamo il valore RD, ovvero, il numero del registro in cui mettere il dato.
A questo punto il valore del registro R1 è aggiornato, e l'istruzione è eseguita.

``` Assembly
LDR R1, [R2, R3]
```
In questa variazione, il campo A1 avrà il valore Rn, ovvero R2. Il campo A2 avrà il valore Rm, ovvero, il registro R3. Il campo A3 avrà il registro R1, come prima.
La ALU sommerà il valore di R2 al valore di R3.


### I Salti
La prima cosa importante da capire è che R15 e PC non sono due registri diversi.
I salti contengono un'etichetta con segno a 24 bit. Dobbiamo estenderlo a 32 bit, mantenendo il segno e moltiplicandolo per quattro, perchè conta il numero di istruzioni.
Il valore che esce dall'estensione deve essere sommato ad otto. Perchè se leggiamo il registro R15 leggiamo PC+8.
Per questo motivo ci sono due sommatori più quattro.

Il register file, almeno nel single cycle, ha i registri da r0 a r14, che, per la lettura, nella porta numero uno hanno un multiplexer. Questo MUX è comandato dai quattro bit dell'indirizzo, ha come input le uscite dei registri, e come output il valore del registro desiderato.
In questo multiplexer, non c'è il registro R15. Il valore che viene restituito dal multiplexer è direttamente il valore PC+8.
In pratica, il registro 15 è un registro fittizio, e la sua lettura dà come risultato il valore del PC aumentato di otto, portato con bypass.
Il valore PC+8 viene portato in uscita. Quindi, facendo una operazione di branch, leggiamo il program counter, che è l'indirizzo dell'istruzione di branch. Nel frattempo PC sta venendo aggiornato.
La ALU somma il valore PC+8 al valore immediato che è stato esteso a 32 bit, e poi questo valore viene scritto di nuovo nel PC.
Ricapitolando: Si fa branch a etichetta, si legge l'indirizzo della istruzione corrente, che è il branch. Si legge l'istruzione dalla memoria istruzioni. L'istruzione di branch è fatta così:

| cond | op | funct | imm24 |
| ---- | --- | ----- | ------ |
| 31:28 | 10 |  1L | 23 : 0 |

La parte controllo capisce che è un salto, allora alla prima porta in lettura dà 15 come valore, e lo forza per leggere r15. La parte controllo dirà di selezionare uno nel commutatore. R15 non è un vero registro, ma si legge attraverso un bypass. Si prende l'immediato che viene esteso e shiftato di due posizioni. L'immediato è sommato a PC+8. Otteniamo il nuovo indirizzo del program counter, che viene sostituito all'indirizzo corrente.

**Qual è il problema del single cycle?**
Il ciclo di clock ha la stessa lunghezza per tutte le istruzioni. Bisogna tarare il ciclo di clock alla istruzione più lunga. Il ciclo di clock considera il massimo ritardo di stabilizzazione.

### Control Unit
Gli ingressi della control unit sono le parti dell'istruzione in linguaggio macchina, ovvero *Cond*, *Op*, *Funct*, e *Rd*, e i flags della ALU.
 Manda in uscita *MemtoReg*, *ALUSrc*, *ImmSrc*, *RegSrc*, *ALUControl*, direttamente. Invece, manda alla conditional logic *FlagW*, *PCS*, *RegW*, *MemW*.

## Architettura multi-cycle
Nell'architettura single-cycle, Fetch, Decode, Execute, Memory, WriteBack, vengono svolti tutti in un singolo ciclo di clock. Questo ci obbliga a tenere il ciclo di clock lungo quanto il tempo per eseguire l'istruzione più lunga: la load. Per la tecnologia del tempo in cui è stato scritto il libro, sono 840 pico-secondi.
Una soluzione potrebbe essere fare il ciclo di clock più piccolo, e svolgere una sola delle operazioni tra fetch, decode ed execute.
Fase di fetch: leggo l'istruzione, prendo pc e lo incremento di quattro, e lo porto in ingresso e lo scrivo.
Secondo ciclo, decode: Quando possiamo iniziare a decodificare l'istruzione, di nuovo usiamo la ALU per aumentare il PC, ma il write enable del program counter sarà zero, perchè ancora non vogliamo aumentare il PC. Però andiamo a portare questo PC+8, nell'ingresso R15 del RF. Quindi nel secondo ciclo siamo già in grado di leggere R15.

**Esempio: store**
```Assembly
STR R0, [R1, #OFF]
```

Primo ciclo: leggiamo il PC e leggiamo l'istruzione e facciamo PC + 4
Secondo ciclo: faccio pc+4+4 e lo porto in R15, nel caso servisse.
Terzo ciclo: Avrò calcolato R1 e l'OFFSET, e utilizzo la ALU per sommarli, e li scrivo in un registro. Nel terzo stadio di execute non si può scrivere da nessuna parte perchè avrò il valore a zero.
Quarto ciclo: scrittura in memoria.

**Esempio: B LOOP**
Primo ciclo: leggiamo un'istruzione all'indirizzo PC, e facciamo PC+4
Secondo ciclo: la parte controllo sa che è una branch, quindi manda al multiplexer un segnale per leggere il registro 15.
Terzo ciclo: l'estensore calcola l'estensione dell'offset quello giusto, abbiamo PC+8, lo sommiamo all'offset. E scriviamo il nuovo valore del PC.

**La parte controllo**
Il decoder è una macchina a stati finiti. Questi stati sono fetch, decode, execute, ma non esattamente. Abbiamo diversi stati execute.
L'automa in questione è un'automa di Moore, cioè le uscite di questo automa dipenderanno solo dallo stato presente. Siamo nello stato di fetch? Manderò dei segnali che dipendono solo dallo stato, e non dalla istruzione.

## Architettura Pipeline
L'idea dell'architettura pipeline è di eseguire più istruzioni allo stesso tempo. Si esegue una istruzione negli stadi fetch, decode, execute, memory read e writeback. Quando la prima istruzione ha eseguito la fetch, passa alla decode, e la seconda istruzione inizia la fetch. E così via.
Quando non ci sono dipendenze, è possibile che ogni istruzione venga semplicemente eseguita un ciclo di clock dopo la precedente.
In modo da implementare cicli multipli per istruzione, ci sono dei registri tra ogni pezzo della DataPath, un registro per ciclo di clock.

