<script
  src="https://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
  type="text/javascript">
</script>

<style>
.markdown-body img {
	max-width: 80%;
	align: center;
}
</style>

# Gerarchia di memoria

## Misure di performance

**Miss Penalty** è il tempo che ci vuole a rimpiazzare un blocco di un livello con con il blocco corrispondente del livello inferiore.
**Miss Time** è i tempo per ottenere un elemento in caso di miss. Il miss time è uguale a miss penalty + hit time.
**Miss Rate** è il rapporto tra numero di miss e il numero totale di accessi alla memoria.
**AMAT** (Average Memory Access Time) è il tempo medio di accesso alla memoria.

$$ AMAT = t_{M0} + MR_{M0} * (t_{M1} + MR_{M1} * (t_{M2} + MR_{M2} * (t_{M3} + ... ) ) ) $$
dove MRx è la miss rate al livello di memoria x, e tx è il tempo di accesso alla memoria al livello x.


### Esempio: effetto della cache su l'AMAT

- *tm = 50ms* (tempo di servizio della memoria principale).
- *tL1 = 1ns* (tempo di servizio nel caso di cache hit).
- Miss rate (*MRL1*): 5%

Nel caso non ci sia cache l'AMAT è uguale a **50ns**, ma con la cache vale solo:

$$ AMAT = t_{L1} + MR_{L1} * t_{M}\ =\ 1+0.05 * 50 = \mathbf{3.5ns} $$.


### Performance della CPU

Il tempo della CPU, detto CPUtime è la moltiplicazione tra IC (instruction count), ovvero il numero di istruzioni eseguite, e CPI (clock cycles per instruction), e la durata del ciclo di clock. CPI è definito come il rapporto tra i cicli di clock e le istruzioni eseguite.

Il tempo della CPU può essere ulteriormente diviso tra i cicli che la CPU spende eseguendo le istruzioni senza cache misses, (*CPI perfect*) e i cicli in cui la CPU aspetta la memoria, *CPI stall*.

$$ CPI_{stall} = \frac{Memory\ accesses}{Program}\ * \ Miss\ Rate\ * \ Miss\ penalty $$

## Memoria cache
Nella memoria cache sono memorizzati i dati dei processi, durante la loro esecuzione. E' molto veloce, infatti il suo tempo di accesso è molto minore al tempo di accesso alla memoria centrale. Un calcolatore può avere due cache contemporaneamente, una per le istruzioni, e una per i dati.
La cache si basa su due concetti fondamentali: la **località temporale** e la **località spaziale**.
> **Località temporale** si accede a locazioni di memoria già utilizzate. 
> **Località spaziale** si accede a locazioni memoria vicine a quelle appena utilizzate.

Ci sono tre tipi di **cache misses**
- *Miss obbligatori*: causati dal primo accesso ad un blocco che non era mai stato nella cache.
- *Miss di capacità*: La cache è troppo grande per contenere tutti i blocchi necessari. Alcuni blocchi sono sostituiti.
- *Miss di conflitto*: causate da un conflitto.

Una cache di capacità **C** è organizzata in **S** sets, ciascuno che contiene **B** blocchi, o righe. **b** piccolo è il numero di parole per blocco. Un'altra caratteristica di una cache è la sua associatività, **N**, o numero di vie.
Quando una cache preleva una parola dalla memoria principale, preleva anche alcune altre parole da aree adiacenti. La dimensione del blocco, *b* indica quante parole ci sono in ogni blocco. `B= C / b`, quindi la capacità è il numero totale di parole.
La relazione tra l'indirizzo di un dato in memoria principale e la sua locazione nella cache è detta **mappatura**. La funzione di mappatura fa corrispondere ad ogni indirizzo di memoria un set della cache.
Nella cache a **mappatura diretta**, ogni set contiene un solo blocco, quindi un indirizzo di memoria viene sempre mappato sulla stessa locazione nella cache. Nella cache **parzialmente associativa** a N vie, ogni set contiene N blocchi, quindi, un indirizzo di memoria viene mappato su un set, ma può venire caricato in uno qualsiasi dei blocchi del set. Nella cache **completamente associativa**, esiste un solo set, quindi una parola di memoria principale può essere mappata in qualsiasi posizione della cache.

### Indirizzi di memoria

| Tag | Set | Block Offset | Byte Offset |
| ---------- | ------- | ------ | -------|
| 001010011001010010111010| 0101101 | 100| 00 |

| Valid | Tag | Data |
| ----------- | ----------- | -------- |
|0 | 001010011001010010111010 | 00101001100101001011101010101010 |

Un dato, in memoria principale, viene descritto da un indirizzo, formato da tag, set, block offset e byte offset.
Il dato, nella cache, viene accompagnato da un bit di validità e dallo stesso tag, che indica la sua posizione in memoria principale.
Il bit **valid** indica se il dato nell'elemento è valido o no. All'avvio del sistema tutti i dati sono marcati come non validi. Il campo **tag** è un valore univoco a 16 bit corrispondente alla linea di memoria da cui provengono i dati. Invece il campo **data** contiene una copia del dato in memoria.

### Cache a mappatura diretta

<img src="DMC.png", alt="direct mapped cache" width="424"/>

Nel contesto della cache a mappatura diretta, gli indirizzi prendono questa forma:

<img src="byte offset.png", alt="address" width="200"/>

**Esempio** Consideriamo una cache con 128 sets, 128 blocchi, e 8 parole per blocco. Qual è la riga di cache e l'offset nel blocco che tiene l'indirizzo 0xFEAC?
- 0xFEAC -> 1111 1110 1010 1100
- Se ci sono 128 sets, cioè 2^7, servono sette bit di set.
- Ci sono otto parole in un blocco, quindi servono 3 bit di block offset.
- Ci sono sempre due bit di byte.
- Quindi (TAG, IDX, OFF) = (1111, 1110101, 01100).
- La riga è quella di indice 117, cioè la 118-esima.
- L'offset nel blocco è 011, cioè tre, quindi è la quarta parola nel blocco.

**Esempio 2**
Consideriamo il seguente blocco di codice
```C
//A, B, C allocati, e A, B inizializzati
for(i=0; i<16; i++) C[i] = A[i] + B[i];
```
Consideriamo un processore che va @2GHz, con una cache a mappatura diretta L1, con C = 128 (contiene massimo 128 parole), otto parole per blocco, quindi ha 16 S, tempo di accesso alla memoria principale tM = 100 cicli, tempo di accesso alla cache tL1 = 4 cicli. 'A' parte all'indirizzo 0x00000000, 'B' starte a 0x000040, e 'C' a 0x0000080. Dopo tM cicli, tutte le parole sono state caricate sul cache block.
Calcoliamo il numero di cache misses, e l'AMAT, considerando solo le istruzioni di load.

- Siccome A inizia all'indirizzo di memoria 0, le prime otto parole di A saranno caricate nel set 0, (000), e le seconde otto nel set 1 (001) della cache L1.
| Array |Tag | Set| Block Offset | Byte Offset|
| ----- |--- | --- | ------------- | ---------- |
| A[0]  | 0...00 | 000 | 000 | 00 |
| A[1]  | 0...00 | 000 | 001 | 00 |
| A[8]  | 0...00 | 001 | 000 | 00 |
| B[0]  | 0...00 | 010 | 000 | 00 |
| C[0]  | 0...00 | 100 | 000 | 00 |

- B sarà caricato nel set 2 (010) e nel set 3 (011) della cache.
- Quindi non ci sono conflitti nei sets, e in totale ci sono quattro misses (due per A e due per B), e 32 cache hits, una per ogni parola nella cache.
- AMAT è uguale a HitTime + MissRate * MissPenalty, e la miss rate è uguale a 4/32, quindi l'amat è uguale a (4 + 1/8 * 100) per la durata del ciclo di clock, dove il ciclo di clock dura 1/2GHz, ovvero 0.5ns, AMAT = 8.25ns.

**Pro e Contro** della Cache a Mappatura Diretta.
- Facile da realizzare,
- Molto veloce in caso di hit,
- Troppo rigida, perchè una parola di memoria viene messa sempre nella stessa riga della cache.

### Caches Associative

La **cache parzialmente associativa** a N vie riduce i conflitti prevedendo N blocchi in ciascun set dove il dato può essere piazzato. Per trovare un determinato blocco, si fa una ricerca in parallelo delle N vie del set, utilizzato N comparatori.

![Cache parzialmente associativa](PAC.png)

La **cache completamente associativa** è costituita da un unico set a B vie, dove B è il numero di blocchi. Questo tipo tende ad avere il numero minore di miss causati da conflitti, ma richiedono hardware aggiuntivo per i confronti del tag.

![Cache completamente associativa](FAC.png)

Aumentare l'associatività N, riduce generalmente il tasso di miss causate da conflitti, ma implica l'uso di comparatori di tag. Aumentare la dimensione di blocco b sfrutta la località spaziale per ridurre il tasso di miss, ma diminuisce il numero di set, quindi può causare più conflitti, e accrescere la penalizzazione di miss.

### Come gestire una miss

Una cache miss blocca l'intero processore, compreso il contenuto dei registri, attendendo la memoria. I passi per gestirla sono:
1. Dire al livello di memoria seguente di leggere il valore
2. Aspettare che la memoria risponda
3. Aggiornare la riga di cache con i dati ricevuti
4. Ricomincia l'esecuzione (stavolta è una hit).

### Gestire le scritture
Modificando un dato, dobbiamo modificare sia il suo valore nella cache che il suo valore nella memoria principale. Ci sono più strategie per fare questo.
Nella strategia **Write through** si aggiornano sia la cache che il livello di memoria successivo. E' una soluzione semplice, e i dati sono sempre consistenti, però la velocità delle writes dipende dalla velocità della memoria di livello inferiore.
Nella strategia **Write back**, una write hit aggiorna solo la cache, poi il blocco modificato viene scritto nella memoria inferiore quando è sostituito. I suoi pro sono che la velocità di scrittura è la stessa della cache, e c'è meno traffico di memoria. I suoi contro sono che dobbiamo tenere traccia dei blocchi modificati con i **dirty bits** e c'è una implementazione più complessa. Inoltre le sostituizioni sono più costose.
Siccome scrivere sulla memoria rigida è costoso, le operazioni di store usano un buffer, chiamato **write buffer**, per tenere i dati che devono essere scritti in memoria.

### Gestire le sostituzioni in cache.
Se nella memoria a mappatura diretta non abbiamo scelta, perchè il blocco può andare in una sola locazione, nelle caches associative abbiamo scelta su dove piazzare il blocco richiesto.
La politica di sostituzione più comune è la LRU (Last Recently Used), per il principio di località temporale, in cui si sostituisce il blocco che non è stato utilizzato da più tempo. Può essere implementato con dei bit U (used), che indicano quale via è stata usata più di recente. Con caches ad alta associatività, una politica random dà una performance equivalente.


