<script
  src="https://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML"
  type="text/javascript">
</script>


# Input Output

## Struttura del sistema I/O
Un device I/O ha due porte, la porta di controllo e la porta dati. Attraverso la porta di controllo possiamo mandare messaggi al dispositivo e ottenere notifiche sullo stato del dispositivo.
In particolare, l'unità di controllo si occupa:
- Della comunicazione con il processore, con la decodifica dei comandi, status reporting, riconoscimento dell'indirizzo.
- Della comunicazione con i device.
- Del buffering dei dati, per ottimizzare i traferimenti e compensare per le differenze di velocità.
- Del riconoscimento degli errori, come quelli di trasmissione(parity bis) e quelli elettro-meccanici.

## Qualità dell'input-output (prestazioni, affidabilità, disponibilità)

**legge di Amdahl**
Parla del massimo speedup ottenibile da un programma parallelo quando si aumenta il numero di processori da 1 a N, ma può essere applicata a qualsiasi processo di ottimizzazione.

*Si consideri un programma di cui solo una frazione f può essere ottimizzata, mentre la frazione (1-f) non viene alterata.*

$$ Speedup\ =\ \frac{Exec.tempo\ prima\ del\ miglioramento}{Exec.tempo\ dopo\ il\ miglioramento} = \frac{T(1-f)+Tf}{T(1-f)+\frac{Tf}{N} } = \frac{1}{(1-f)+\frac{f}{N}} $$

Lo speedup è limitato dalla frazione sequenziale (1-f), ovvero dalla parte del processo che non posso migliorare!

### Affidabilità e disponibilità
L'affidabilità di un sottosistema I/O si misura con *Mean time to Failure* (MTTF), e si ottiene con la *fault avoidance*(prevenzione degli errori, e la *fault tolerance*(rindondanza). La disponibilità di un sottosistema si misura con il *Mean time to repair*(MTTR), infatti il suo valore è pari a MTTF/ MTTF * MTTR.

### Connessione bus
Un bus è una collezione di linee di dato, che sono trattate come un singolo segnale logico. Un bus è un mezzo a trasmissione condivisa, cioè un solo dispositivo può usarlo alla volta. Il **system bus** connette tutti i componenti principali, ed è composto da linee separate in tre gruppi: data, indirizzo, controllo.

Le comunicazioni bus devono essere gestite con un protocollo di comunicazione. Si dice **bus master** un'unità che può iniziare una richiesta di utilizzo. E' necessario scegliere un master tra multiple richieste, cercando di implementare la priorità e la giustizia, prevenendo la starvation. Esistono modelli di arbitrazioni centralizzati e condivisi.

Nell'arbitrazione **daisy-chain**, i devices si connettono al bus in ordine di priorità - i dispositivi ad alta priorità intercettano e negano le richieste dei dispositivi di priorità inferiore. E' una soluzione centralizzata che è semplice da implementare, ma è lenta e non garantisce equità.

### Gestione I/O

**In che modo la CPU dà comandi ai device?**
Il sistema operativo è l'unico a poter dare comandi al device I/O, e se un programma a livello utente vuole interagire deve utilizzare le *system calls*. Ci sono due metodi principali per inviare comandi ai devices:
- Istruzioni I/O
- I/O memory-mapped.

Le istruzioni I/O sono istruzioni ISA che alterano i registri del device. Le istruzioni privilegiate sono disponibili solo in kernel mode.

Nella memory-mapped I/O, una porzione dello spazio fisico di memoria è riservato all'I/O. I registri interni del device sono mappati su locazioni riservate della memoria principale. I comandi I/O sono read/write standard. Le operazioni sono ridirette ai controllers tramite la Memory Management Unit.

**Esempio: memory-mapped I/O**
Supponiamo che un I/O device sia assegnato l'indirizzo di memoria 0x20001000. Mostriamo il codice Assembly ARM per scrivere il valore 7 nel device, e per leggere il valore di output dal device.

```Assembly
	MOV r1, #7
	LDR r2, =ioadr
	STR r1, [R2]
ioadr: DCD 0x20001000
```
```Assembly
	LDR r1, [r2]
```

**Come fa la CPU a sapere quando i devices hanno finito l'operazione?**
Nel caso in cui l'I/O sia programmata, la CPU controlla direttamente le operazioni e lo status del controllo. Il device ha un ruolo passivo, e la CPU attende che il modulo abbia completato le operazioni ciclicamente, nell'operazione nota come polling. Esaminando ciclicamente il registro di stato del device, la CPU spreca moltissimi cicli, perchè potrebbe leggerlo molte volte per scoprire che il device non ha completato l'operazione.

Una I/O potrebbe essere anche **interrupt driven**. Un *interrupt* è un evento asincrono che viene da un device. Le interruzioni sono un'alternativa al polling: la CPU dà comandi al device e continua a fare altre tasks, e il dispositivo I/O genera interruzioni quando lo stato cambia. Le interruzioni I/O sono asincrone e divise per priorità.

### Trasferimento dati e Direct Memory Access
Il **Direct Memory Access** o DMA è un meccanismo che dà ad un device I/O l'abilità di trasferire dati da e verso la memoria centrale, senza coinvolgere la CPU.
Il DMA separa il protocollo di comunicazione tra CPU e memoria dal protocollo fra memoria e device I/O. Le interruzioni sono usate dal device per comunicare con la CPU solo al completamento del trasferimento, o quando avviene un errore. Il controllo DMA è speciale: può essere condiviso tra più device, ed è esso stesso visto come un dispositivo I/O memory-mapped. Si occupa di arbitrazione bus e trasferimento dati. Il controllo DMA e la CPU si contendono il memory bus.

**Trasferimento DMA** il trasferimento DMA ha tre fasi:
1. La CPU dà al DMA le informazioni di identità, operazione da eseguire, indirizzo di memoria e numero di bytes da trasferire.
2. Il DMA esegue l'operazione sul device e arbitra la connessione. Trasferisce i dati.
3. Quando il trasferimento è finito, il controllo invia un'interruzione alla CPU.

Esistono DMA fisiche e virtuali. Il controllo del DMA virtuale deve effettuare l'address translation internamente, utilizzando una piccola cache (TLB), inizializzata dal sistema operativo. Il controllo del DMA fisico lavora con indirizzi fisici, e trasferisce seguendo la granularità di pagina, infatti il SO divide i trasferimenti larghi in pezzi della dimensione di una pagina. Il DMA fisico è più semplice e meno flessibile del DMA virtuale.

## Device drivers
Il sistema operativo definisce un name-space per i dispositivi e i loro identificatori, implementa le politiche di cache ed è indipendente dall'hardware. Il **device driver** lavora in kernel space, si interfaccia con il device controller, si occupa della sincronizzazione, si interfaccia con il DMA controller per occuparsi del trasferimento di dati, e gestisce gli errori del dispositivo.

