# Memoria

La gestione della memoria si occupa di allocare la memoria fisica ai processi che ne fanno richiesta. Nessun processo può diventare attivo prima di aver ottenuto uno spazio di memoria.
Un buon gestore di memoria in un ambiente multi-programmato deve garantire la protezione dei dati, impedendo ai processi attivi di sconfinare nello spazio di indirizzamento di altri processi. Il gestore di memoria, allo stesso tempo, deve permettere la condivisione di dati, perchè i processi che cooperano possano accedere ad aree comuni di memoria. I moduli di gestione della memoria si differenziano in base al tipo di allocazione della stessa, che può essere contigua o non contigua.

## Allocazione contigua

La memoria viene allocata in modo tale che ciascun oggetto occupa un insieme di locazioni i cui indirizzi sono strettamente consecutivi.
Esistono vari tipi di allocazione contigua:
- Monoallocazione
- Partizionamento statico
- Partizionamento dinamico
- Segmentazione

### Monoallocazione
La **monoallocazione** è lo schema più semplice per la gestione di memoria che viene suddivisa in due aree contigue:
- la prima allocata permanentemente ad una porzione residente del sistema operativo.
- La seconda assegnata ai processi in esecuzione, per il tempo necessario al completamento.

Il sistema operativo:
- Tiene traccia della prima e dell'ultima locazione disponibile per i processi non permanenti.
- Viene posto ad uno dei due estremi della memoria
- Contiene i vettori di interrupt (interrupt table)
Altri moduli temporanei del sistema operativo sono all'altra estremità della memoria.

Alla richiesta di esecuzione di un programma, il sistema operativo si assicura che le dimensioni del processo siano compatibili con la memoria disponibile, conferisce il controllo al processo utente fino al suo completamento o ad eventuali condizioni di errore. Al termine del processo la memoria viene liberata e può essere assegnata ad un altro processo in attesa.
Raramente un monitor monoprocesso supporta meccanismi di protezione tra processi utente in quanto in ogni istante si può tenere un solo processo in memoria. Però può avere meccanismi di protezione del codice del sistema operativo da sconfinamenti del processo.
La protezione del sistema operativo è effettuata attraverso supporti hardware come:
- Registri barriera
- Diritti di accesso mediante bit di protezione
- Sistema operativo in memoria di sola lettura.

I *registri barriera* sono usati per tracciare un confine tra aree dei processi di sistema e dei processi. Nel registro viene memorizzata la prima locazione disponibile al processo e il sistema operativo effettua i controlli di sconfinamento. Questo metodo richiede di distinguere l'esecuzione del sistema operativo e dei processi.
Un altro metodo di progettare la memoria consiste nei *diritti di accesso*, mediante bit di protezione, nella memoria stessa. Ad ogni parola di memoria viene associato un bit di protezione che viene posto a 1 nelle zone che contengono il sistema operativo, e a 0 nelle restanti. I processi utente accedono solo a parole con bit 0, mentre il sistema operativo ha accesso illimitato.
Il metodo di mettere il sistema operativo in una *memoria di sola lettura* non è solitamente utilizzato per la sua scarsa flessibilità e impossibilità di aggiornare il codice.

La memoria condivisa non ha senso negli ambienti monoprocesso, e raramente viene supportata dai relativi monitor. I programmi utente potrebbero passarsi dati ponendoli in locazioni di memoria che non vengono sovrascritte durante l'esecuzione dei processi.

La monoallocazione rende il sistema meno efficiente, perchè la memoria può essere sovradimensionata per alcuni processi, e troppo piccola per altri. D'altra parte il costo della progettazione è più basso.

### Partizionamento statico
Il **partizionamento statico** è un metodo per supportare la multiprogrammazione, e consiste di dividere la memoria in partizioni, ciascuna delle quali può essere allocata a un processo diverso. Il partizionamento statico implica che la suddivisione della memoria venga fatta prima dell'esecuzione e che da quel momento le partizioni rimangono di dimensioni fisse. Il numero e la dimensione di ciascuna partizione vengono determinati durante la generazione del sistema considerando:
- Capacità della memoria fisica disponibile
- Grado desiderato di multiprogrammazione
- Dimensione dei processi più frequentemente eseguiti.
Il sistema operativo tiene traccia delle partizioni in un documento chiamato tabella di descrizioni delle partizioni (TDP).
Talvolta non è possibile allocare un processo, perchè tutte le partizioni sono allocate, o non ci sono partizioni di memoria sufficientemente grandi.
E' possibile dividere il programma di origine in sottoprogrammi. Lo scheduler e il gestore della memoria lavorano insieme: lo scheduler determna queli processi sono pronti, e quindi in memoria. Il gestore può decidere di sfrattare un processo in memoria per liberare una partizione.
Lo **swapping** è l'operazione di rimozione dalla memoria di un processo e il suo successivo caricamento. Lo swapping ha un costo sul carico di lavoro della CPU.
Quando lo scheduler decide l'introduzione in memoria di un nuovo processo, il gestore sceglie uno da rimuovere, in base a
- dimensione della partizione richiesta
- priorità del processo
- tipo di evento atteso dal processo
- tempo trascorto in memoria.
Un processo che è già parzialmente eseguito è costituito da codice eseguibile, dati già elaborati, stack, registri di stato, file aperti e metadati.
Al momento di swapping questi oggetti sono registrati in un file su hard disk, detto file di swapping. Uno swapping efficiente richiede che i processi siano reallocabili dinamicamente.

### Partizionamento dinamico
Nel *partizionamento dinamico* il sistema delle partizioni è analogo a quello del partizionamento statico. Il gestore della memoria crea e assegna partizioni di dimensione variabile dinamicamente in base alle richieste dei processi. Quando un processo termina subisce uno swapping, e il gestore restituisce il suo spazio alla lista libera.
Lo schema di partizionamento dinamico genera buchi di memoria, nel fenomeno detto *frammentazione*. Quando avviene, è necessario effettuare la compattazione della memoria che consiste nello spostare i processi residenti in memoria per creare partizioni libere più grandi. E' una operazione ad alto costo, che può essere svolta continuamente o su necessità.

### Segmentazione
La *segmentazione* è uno schema di memoria che si basa sulla divisione dello spazio di indirizzamento dei processi in entità logiche che possono essere poste in aree non contigue di memoria. Fornisce strumenti per la reallocazione dinamica, la protezione e la condivisione. Un programma eseguibile può essere suddiviso, ad esempio, in codice, dati e stack: ciascuno di questi può essere posto in un segmento diverso di dimensioni differenti.
> Un unico segmento deve essere in una locazione di memoria contigua.
L'indirizzo nella segmentazione è bidimensionale, contiene il nome del segmento e l'offset al suo interno.
Il caricamento di un processo segmentato avviene in questo modo:
- Il SO cerca di allocare n partizioni per ciascun segmento.
- Il SO crea un descrittore di segmento, che indica l'indirizzo fisico e la dimensione.
- L'insieme dei descrittori di segmento viene raccolto nella TDS, tabella dei descrittori di segmento.
Le dimensioni della TDS sono legale alle dimensioni dello spazio virtuale di indirizzamento di un processo. La traduzione di ciascun indirizzo virtuale bidimensionale richiede due accessi in memoria, uno alla TDS per leggere l'indirizzo fisico, e l'altro all'indirizzo trovato. Quindi, la segmentazione **raddoppia** il tempo per accedere ad una locazione di memoria.

